#include <iostream>
#include <memory>
#include <cmath>
#include <algorithm>
#include "WiiCmd.h"
#include "Draw.h"

#include "gfx/wiimote_img.h"
#include "gfx/wiimote_up.h"
#include "gfx/wiimote_axis.h"
#include "gfx/bg.h"

#include "GRRLIB/GRRLIB.h"
Mtx GXmodelView2D;

using namespace std;

enum states{
	START,
	CALIBRATING_ZERO,
	CALIBRATED_ZERO,
	CALIBRATING_MAX,
	CALIBRATED,
	CALIBRATING_FREEZING,
};


void discardEmptyFrames(){
	WiiCmd* w = WiiCmd::getInstance();
	Draw* d = Draw::getInstance();
	
	while( 
		(w->getAccel().x == 0) and
		(w->getAccel().y == 0) and
		(w->getAccel().z == 0)
	){
		//w->waitVSync();
		d->render();
		w->update();
		if ( w->isDown(WPAD_BUTTON_HOME) ){ exit(0); }
	}
}

int main(int argc, char **argv){

	const string version = "0.5";
	const int calibframes = 1 * 60;
	const double pi = 3.1415;
	const string compiledate(__DATE__);
	const string compiletime(__TIME__);
	const double diffthreshold = 2;
	
	double ref_x = 0;
	double ref_y = 0;
	double ref_z = 0;
	double drawdelta_x = 0;
	double drawdelta_y = 0;
	double drawdelta_z = 0;
	double delta_x = 0;
	double delta_y = 0;
	double delta_z = 0;			
	double delta_max = 1; //the maximum value that can result in delta_{x,y,z}, so we can convert it to degrees later - calibrated in CALIBRATING_MAX
	bool isFrozen = false;
	bool showDebug = false;
	states state = START;

	WiiCmd* w = WiiCmd::getInstance();
	Draw* d = Draw::getInstance();
	//auto_ptr<Draw> d(new Draw);
	
	//Load pictures
	u8* tex_wmaxis = d->getTextureFromBuffer(wiimote_axis); //296x248
	u8* tex_wmup = d->getTextureFromBuffer(wiimote_up); //200x400
	u8* tex_wmimg = d->getTextureFromBuffer(wiimote_img); //400x132
	u8* tex_bg = d->getTextureFromBuffer(bg); //640x480
	
	//We do not know what kind of data is in our button&co memory at the beginning,
	//so -in order to not quit immediately- we need to fill it here.
	w->update();

	while(!w->isDown(WPAD_BUTTON_HOME)) {
		
		d->render();
		
		d->setColor(C_BLACK);
		//d->fillScreen();
		d->drawTexture(0,0,640,480,tex_bg);
		
		w->update();
			
		//Drawing begins here
		if(state == START){
			d->setColor(C_BLACK);
			d->drawTexture(120,270,400,132,tex_wmimg,true);
			
			d->setColor(C_WHITE);
			d->drawText(3,"This is Leveltool "+ version +".");
			
			if(!w->wiimoteConnected(0)){
				d->drawText(5,"  Waiting for Wiimote #1 to connect.");
			}
			else{
				d->drawText(5,"  Please put your Wiimote #1 on a reference surface and press A.");
				d->drawText(6,"  This will begin a calibration to find your wiimote's 'zero' level.");
				d->drawText(7,"  After this, another calibration will be required to find the maximum.");
				
				d->drawText(9,"Press A to continue.");
				d->drawText(10,"Press HOME to quit.");
				if ( w->isDown(WPAD_BUTTON_A) ){ state = CALIBRATING_ZERO; }
			}
		}
		
		if(state == CALIBRATING_ZERO){
			//Make sure that these values are zero, because this state can be called again, and we don't want the old values to pollute our new ones.
			ref_x = 0;
			ref_y = 0;
			ref_z = 0;
			drawdelta_x = 0;
			drawdelta_y = 0;
			drawdelta_z = 0;

			//Sometimes we start out with zeroes for our accel values, which would mess up the
			//rest down below. To fix that, we skip these frames.
			discardEmptyFrames();
			
			int x[calibframes];
			int y[calibframes];
			int z[calibframes];
			//Here we collect data to get meaningful reference values
			for(int framecounter = 0; framecounter < calibframes; framecounter++){
				d->drawTexture(0,0,640,480,tex_bg);
				d->drawText(3,"Calibrating zero level.");
				d->render();
				w->update();
				if ( w->isDown(WPAD_BUTTON_HOME) ){ exit(0); }
				
				x[framecounter] = w->getAccel().x;
				y[framecounter] = w->getAccel().y;
				z[framecounter] = w->getAccel().z;
			}
			//Calc' the median, more resistant to 'static'
			sort(x,x+calibframes);
			sort(y,y+calibframes);
			sort(z,z+calibframes);
			ref_x = x[calibframes/2];
			ref_y = y[calibframes/2];
			ref_z = z[calibframes/2];	
			
			state = CALIBRATED_ZERO;
		}
		
		if(state == CALIBRATED_ZERO){
			d->setColor(C_BLACK);
			d->drawTexture(400,40,200,400,tex_wmup,true);
			d->setColor(C_WHITE);
			d->drawText(3,"Now, please put your wiimote upright");
			d->drawText(4,"and press A again.");
				
			d->drawText(6,"Press HOME to quit.");
			if ( w->isDown(WPAD_BUTTON_A) ){ state = CALIBRATING_MAX; }
		}
		
		if(state == CALIBRATING_MAX){
		
			//d->drawText(3,"Waiting for acceleration values.");
			discardEmptyFrames();
			
			//Here we determine the Wiimotes maximum value		
			//Z will report a too low value if standing upright, so we skip it.
			//...and X will report a wrong value if the Wiimote was calibrated
			//with the button side on top.
			double y[calibframes];
			for(int framecounter = 0; framecounter < calibframes; framecounter++){
				d->drawTexture(0,0,640,480,tex_bg);
				d->drawText(3,"Calibrating maximum level.");
				d->render();
				w->update();
				if ( w->isDown(WPAD_BUTTON_HOME) ){ exit(0); }
				
				y[framecounter] = ((double) abs(w->getAccel().y - (int)ref_y));
			}
			sort(y,y+calibframes);
			
			delta_max = y[calibframes/2] - 2; //we subtract some off the maximum to reduce flickering when getting near 90 degrees
			isFrozen = false;
			state = CALIBRATED;
		}

		if(state == CALIBRATED){
			if(!isFrozen){
				delta_x = (double) w->getAccel().x - ref_x;
				delta_y = (double) w->getAccel().y - ref_y;
				delta_z = (double) w->getAccel().z - ref_z;
			}

			
			//If our delta flickers only about 1 unit, we won't move the pos box
			double diffx = delta_x - drawdelta_x;
			double diffy = delta_y - drawdelta_y;
			double diffz = delta_z - drawdelta_z;
			
			if( (diffx > diffthreshold) or (diffx < -diffthreshold) ){ drawdelta_x = delta_x; }
			if( (diffy > diffthreshold) or (diffy < -diffthreshold) ){ drawdelta_y = delta_y; }
			if( (diffz > diffthreshold) or (diffz < -diffthreshold) ){ drawdelta_z = delta_z; }
			
			//Convert to degrees
			//Limits to prevent values from going to high
			double norm_x = drawdelta_x/delta_max; if(norm_x > 1){ norm_x = 1; }; if(norm_x < -1){ norm_x = -1; }
			double norm_y = drawdelta_y/delta_max; if(norm_y > 1){ norm_y = 1; }; if(norm_y < -1){ norm_y = -1; }
			double norm_z = drawdelta_z/delta_max; if(norm_z > 1){ norm_z = 1; }; if(norm_z < -1){ norm_z = -1; }
			
			double angle_x = (asin(norm_x)*180.0) / pi;
			double angle_y = -(asin(norm_y)*180.0) / pi;
			double angle_z = -(asin(norm_z)*180.0) / pi;
			
			d->setColor(C_BLACK);
			d->fill50Box(319,95,305,257);
			d->drawTexture(324,100,296,248,tex_wmaxis);
			
			d->setColor(C_WHITE);
			d->drawText(3,"You can now place the Wiimote on the target surface.");
			
			d->drawText(5,"Difference from reference surface in degrees:");
			
			if(showDebug){
				const int dbg_x = 20;
				const int dbg_y = 240;
				
				d->setColor(C_BLACK);
				d->fill50Box(dbg_x-5,dbg_y-5,380,150);
				d->setColor(C_WHITE);
				d->drawFreeText(dbg_x,dbg_y+14*0,"- Calibration and drawing -");
				d->drawFreeText(dbg_x,dbg_y+14*1,"X0: "+ d->intToString(ref_x) +", X: "+ (d->intToString(delta_x+ref_x)) +" ("+ d->intToString(delta_x) +"), draw diff.: "+ d->intToString(drawdelta_x) +"    " );
				d->drawFreeText(dbg_x,dbg_y+14*2,"Y0: "+ d->intToString(ref_y) +", Y: "+ (d->intToString(delta_y+ref_y)) +" ("+ d->intToString(delta_y) +"), draw diff.: "+ d->intToString(drawdelta_y) +"    " );
				d->drawFreeText(dbg_x,dbg_y+14*3,"Z0: "+ d->intToString(ref_z) +", Z: "+ (d->intToString(delta_z+ref_z)) +" ("+ d->intToString(delta_z) +"), draw diff.: "+ d->intToString(drawdelta_z) +"    " );
				d->drawFreeText(dbg_x,dbg_y+14*4,"Max. diff.: "+ d->intToString(delta_max) +", Frozen: "+ d->boolToString(isFrozen));
				
				d->drawFreeText(dbg_x,dbg_y+14*6,"- Constants -");
				d->drawFreeText(dbg_x,dbg_y+14*7,"Compiled: "+ compiledate + ", "+ compiletime +". Version: "+ version +"");
				d->drawFreeText(dbg_x,dbg_y+14*8,"Calibration frames: " + d->intToString(calibframes) );
				d->drawFreeText(dbg_x,dbg_y+14*9,"Draw threshold: "+ d->doubleToString(diffthreshold) );
			}
			else{
				//Draw an arc-graph showing the angle
				const double arc_magnifier = 73.0;
				const int arcbegin_x = 100;
				const int arcbegin_y = 295;
				double yarc_x = cos( (-angle_y*pi)/180.0 ) * arc_magnifier;
				double yarc_y = sin( (-angle_y*pi)/180.0 ) * arc_magnifier;
				double zarc_x = cos( (-angle_z*pi)/180.0 ) * arc_magnifier;
				double zarc_y = sin( (-angle_z*pi)/180.0 ) * arc_magnifier;
				d->setColor(C_GREY);
				d->drawLine(arcbegin_x,arcbegin_y,arcbegin_x+arc_magnifier,arcbegin_y); //baseline x
				d->drawLine(arcbegin_x,arcbegin_y-arc_magnifier,arcbegin_x,arcbegin_y+arc_magnifier); //baseline y
				d->setColor(C_BLUE);
				d->drawLine(arcbegin_x,arcbegin_y,(int) arcbegin_x+yarc_x,(int) arcbegin_y+yarc_y);
				d->setColor(C_GREEN);
				d->drawLine(arcbegin_x,arcbegin_y,(int) arcbegin_x+zarc_x,(int) arcbegin_y+zarc_y);
				d->setColor(C_WHITE);
				d->drawFreeText(10,arcbegin_y-7,"Graph:");
				d->drawFreeText(arcbegin_x-8,arcbegin_y-arc_magnifier-14,"90");
				d->drawFreeText(arcbegin_x-16,arcbegin_y+arc_magnifier,"-90");
				d->drawFreeText(arcbegin_x+arc_magnifier,arcbegin_y-8,"0");
			}
			
			const int box_x = 100;
			const int box_y = 108;
			const int box_w = 180;
			const int box_h = 20;
			const int box_space = 32;
			const int box_textoffset_x = -60;
			
			d->setColor(C_BLACK);
			d->fill50Box(box_x+box_textoffset_x-5,box_y-5,box_w-box_textoffset_x+10,box_h+5+5+box_space+box_space);
			
			d->setColor(C_WHITE);
			d->drawFreeText(box_x+box_textoffset_x,box_y+3,"X: "+ d->intToString((int)angle_x) );
			d->drawFreeText(box_x+box_textoffset_x,box_y+box_space+3,"Y: "+ d->intToString((int)angle_y) );
			d->drawFreeText(box_x+box_textoffset_x,box_y+box_space+box_space+3,"Z: "+ d->intToString((int)angle_z) );
			
			d->drawSliderBox(box_x,box_y,box_w,box_h,((int)angle_x)+90 );
			d->drawSliderBox(box_x,box_y+box_space,box_w,box_h,((int)angle_y)+90 );
			d->drawSliderBox(box_x,box_y+box_space+box_space,box_w,box_h,((int)angle_z)+90 );
			
			d->drawText(30,"Press A to recalibrate, B to toggle debug info,");
			d->drawText(31,"(+) to (un)freeze the current values or HOME to exit.");

			if ( w->isDown(WPAD_BUTTON_A) ){ state = CALIBRATING_ZERO; }
			if ( w->isDown(WPAD_BUTTON_B) ){ showDebug ? showDebug = false : showDebug = true; }
			if ( w->isDown(WPAD_BUTTON_PLUS) ){ isFrozen ? isFrozen = false : state = CALIBRATING_FREEZING; }
		}
		
		if(state == CALIBRATING_FREEZING){
			discardEmptyFrames();
			
			int x[calibframes];
			int y[calibframes];
			int z[calibframes];
			for(int framecounter = 0; framecounter < calibframes; framecounter++){
				d->drawTexture(0,0,640,480,tex_bg);
				d->drawText(3,"Freezing.");
				d->render();
				w->update();
				if ( w->isDown(WPAD_BUTTON_HOME) ){ exit(0); }
				
				x[framecounter] = w->getAccel().x - ref_x;
				y[framecounter] = w->getAccel().y - ref_y;
				z[framecounter] = w->getAccel().z - ref_z;
			}
			sort(x,x+calibframes);
			sort(y,y+calibframes);
			sort(z,z+calibframes);
			delta_x = x[calibframes/2];
			delta_y = y[calibframes/2];
			delta_z = z[calibframes/2];
			isFrozen = true;
			state = CALIBRATED;
		}

	}
	
	//Cleanup. I learned the hard way that leaving your memory around on the Wii can be frustrating.
	delete w;
	ref_x = 0;
	ref_y = 0;
	ref_z = 0;
	drawdelta_x = 0;
	drawdelta_y = 0;
	drawdelta_z = 0;
	delta_max = 0;
	delta_x = 0;
	delta_y = 0;
	delta_z = 0;			
	isFrozen = false;
	
	return 0;
}
